#include "base64.hpp"

size_t base64::decoded_length(const void * input, const size_t length) throw() {
	if (length > 0 && static_cast<const char *>(input)[length - 1] == '=') {
		if (length > 1 && static_cast<const char *>(input)[length - 2] == '=') {
			return ((length / 4) * 3) - 2;
		}
		return ((length / 4) * 3) - 1;
	}
	return (length / 4) * 3;
}

size_t base64::decode(void * output, const void * input, const size_t length) throw() {
	// This maps provides efficient conversion between the ascii and base64 character set.
	// Conversion is done by subtracting 43 from characters and performing index loopups.
	static const uint8_t map[128] = {
		62, // "+"
		0, 0, 0,
		63, // "/"
		52, 53, 54, 55, 56, 57, 58, 59, 60, 61, // 0-9
		0, 0, 0, 0, 0, 0, 0,
		0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, // A-Z
		0, 0, 0, 0, 0, 0,
		26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, // a-z
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0,
	};
	const uint8_t * in = static_cast<const uint8_t *>(input);
	uint8_t * out = static_cast<uint8_t *>(output);
	for (size_t i = 0, end = length / 4; i < end; i++) {
		const uint8_t c1 = map[((*in++) - 43) & 0x7f];
		const uint8_t c2 = map[((*in++) - 43) & 0x7f];
		const uint8_t c3 = map[((*in++) - 43) & 0x7f];
		const uint8_t c4 = map[((*in++) - 43) & 0x7f];
		const uint32_t block = (c1 << 18) | ( c2 << 12) | (c3 << 6) | (c4);
		*out++ = (block >> 16) & 0xff;
		*out++ = (block >> 8) & 0xff;
		*out++ = block & 0xff;
	}
	return decoded_length(input, length);
}

size_t base64::encode(void * output, const void * input, const size_t length) throw() {
	// This maps provides efficient conversion between the base64 and ascii character set.
	static const char map[64] = {
		'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P',
		'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', 'a', 'b', 'c', 'd', 'e', 'f',
		'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v',
		'w', 'x', 'y', 'z', '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '+', '/'
	};
	const uint8_t * in = static_cast<const uint8_t *>(input);
	char * out = static_cast<char *>(output);
	for (size_t i = 0, end = length / 3; i < end; i++) {
		*out++ = map[in[0] >> 2];
		*out++ = map[((in[0] & 0x3) << 4) | (in[1] >> 4)];
		*out++ = map[((in[1] & 0xf) << 2) | in[2] >> 6];
		*out++ = map[in[2] & 0x3f];
		in += 3;
	}
	switch (length % 3) {
		case 1: {
			*out++ = map[in[0] >> 2];
			*out++ = map[(in[0] & 0x3) << 4];
			*out++ = '=';
			*out++ = '=';
			break;
		case 2: {
			*out++ = map[in[0] >> 2];
			*out++ = map[((in[0] & 0x3) << 4) | (in[1] >> 4)];
			*out++ = map[(in[1] & 0xf) << 2];
			*out++ = '=';
			break;
		}
		}
	}
	return encoded_length(length);
}

bool base64::valid(const void * input, const size_t length) throw() {
	const char * in = static_cast<const char *>(input);
	if (length % 4) { return false; }
	size_t l = length;
	while (l && in[l - 1] == '=') { l--; }
	for (size_t i = 0; i < l; i++) {
		const char c = *in++;
		if (!(
			(c >= 'A' && c <= 'Z') ||
			(c >= 'a' && c <= 'z') ||
			(c >= '0' && c <= '9') ||
			(c == '+') ||
			(c == '/')
		)) {
			return false;
		}
	}
	return true;
}
