#ifndef __FILE_HPP__
#define __FILE_HPP__

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <cstring>
#include "globals.hpp"
#include "config.h"
#include "util.hpp"

class file_t {
public:

	file_t(
		const std::string & filename,
		const int flags);

	file_t(
		const std::string & filename,
		const int flags,
		mode_t mode);

	file_t(file_t && handle) throw();

	file_t(const file_t & handle) = delete;

	~file_t();

	void close();

	time_t modified() const;

	const std::string & name() const throw();

	void read(
		void * output,
		const ssize_t size);

	void rename(const std::string & to);
	static void rename(
		const std::string & from,
		const std::string & to);

	off_t size() const;

	static file_t temporary();
	static file_t temporary(const std::string & basename);

	ssize_t try_read(void * output, const ssize_t size);

	void write(
		const void * input,
		const ssize_t size);

	file_t & operator=(file_t && handle) throw();

	file_t & operator=(const file_t & handle) = delete;

private:

	const std::string m_name;
	int m_handle;

	file_t(
		const int handle,
		const std::string & filename);
};

inline file_t::file_t(
	const std::string & filename,
	const int flags
) : m_name(filename)
, m_handle(open(filename.data(), flags))
{
	if (m_handle == -1) { throw_system_error(); }
}

inline file_t::file_t(
	const std::string & filename,
	const int flags,
	mode_t mode
) : m_name(filename)
, m_handle(open(filename.data(), flags, mode))
{
	if (m_handle == -1) { throw_system_error(); }
}

inline file_t::file_t(
	const int handle,
	const std::string & filename
) : m_name(filename)
, m_handle(handle) {
	if (m_handle == -1) { throw_system_error(); }
}

inline file_t::~file_t() {
	close();
}

inline void file_t::close() {
	if (m_handle != -1) {
		::close(m_handle);
	}
}

inline time_t file_t::modified() const {
	struct stat info;
	if (fstat(m_handle, &info) != 0) { throw_system_error(); }
	return info.st_mtime;
}

inline const std::string & file_t::name() const throw() {
	return m_name;
}

inline void file_t::read(
	void * output,
	const ssize_t size
) {
	if (::read(m_handle, output, size) != size) { throw_system_error(); }
}

inline void file_t::rename(const std::string & to) {
	rename(m_name, to);
}

inline file_t file_t::temporary() {
	return temporary(XLIC_TMP);
}

inline off_t file_t::size() const {
	struct stat info;
	if (fstat(m_handle, &info) != 0) { throw_system_error(); }
	return info.st_size;
}

inline ssize_t file_t::try_read(
	void * output,
	const ssize_t size
) {
	return ::read(m_handle, output, size);
}

inline void file_t::write(
	const void * input,
	const ssize_t size
) {
	if (::write(m_handle, input, size) != size) { throw_system_error(); }
}

#endif
