#ifndef __LICENSE_KEY_HPP__
#define __LICENSE_KEY_HPP__

#include <cstdint>
#include <cstring>
#include <xorcom/licensing.h>
#include "base64.hpp"
#include "hex.hpp"
#include "config.h"

class license_key_t : public xlic_license_key {
public:

	static const license_key_t invalid;

	license_key_t();
	explicit license_key_t(const xlic_license_key & key) throw();

	static license_key_t from_base64(const char * input) throw();
	static license_key_t from_base64(
		const void * input,
		const size_t length) throw();

	static license_key_t from_hex(const char * input) throw();
	static license_key_t from_hex(
		const void * input,
		const size_t length) throw();

	std::string to_base64() const;

	std::string to_hex() const;

	xlic_license_type type() const throw();

	license_key_t & operator=(const xlic_license_key & key) throw();

	bool operator==(const license_key_t & key) const throw();
	bool operator==(const xlic_license_key & key) const throw();

	bool operator!=(const license_key_t & key) const throw();
	bool operator!=(const xlic_license_key & key) const throw();
};

inline license_key_t::license_key_t()
: xlic_license_key(invalid) {}

inline license_key_t::license_key_t(const xlic_license_key & key) throw()
: xlic_license_key(key) {}

inline license_key_t license_key_t::from_base64(const char * input) throw() {
	return from_base64(input, strlen(input));
}
inline license_key_t license_key_t::from_base64(
	const void * input,
	const size_t length
) throw() {
	if (base64::decoded_length(input, length) == sizeof(xlic_license_key) && base64::valid(input, length)) {
		license_key_t tmp;
		base64::decode(tmp.data, input, length);
		return tmp;
	}
	return invalid;
}

inline license_key_t license_key_t::from_hex(const char * input) throw() {
	return from_hex(input, strlen(input));
}

inline license_key_t license_key_t::from_hex(
	const void * input,
	const size_t length
) throw() {
	if (hex::decoded_length(length) == sizeof(xlic_license_key) && hex::valid(input, length)) {
		license_key_t tmp;
		hex::decode(tmp.data, input, length);
		return tmp;
	}
	return invalid;
}

inline std::string license_key_t::to_base64() const {
	return base64::encode(data, sizeof(xlic_license_key));
}

inline std::string license_key_t::to_hex() const {
	return hex::encode(data, sizeof(xlic_license_key));
}

inline xlic_license_type license_key_t::type() const throw() {
	if (operator==(invalid)) {
		return XLIC_INVALID_LICENSE;
	}
	return (data[2] & (0x08)) ?  XLIC_PHYSICAL_LICENSE : XLIC_VIRTUAL_LICENSE;
}

inline license_key_t & license_key_t::operator=(const xlic_license_key & key) throw() {
	memcpy(data, key.data, sizeof(xlic_license_key));
	return *this;
}

inline bool license_key_t::operator==(const license_key_t & key) const throw() {
	return (memcmp(data, key.data, sizeof(xlic_license_key)) == 0);
}

inline bool license_key_t::operator==(const xlic_license_key & key) const throw() {
	return (memcmp(data, key.data, sizeof(xlic_license_key)) == 0);
}

inline bool license_key_t::operator!=(const license_key_t & key) const throw() {
	return (memcmp(data, key.data, sizeof(xlic_license_key)) != 0);
}

inline bool license_key_t::operator!=(const xlic_license_key & key) const throw() {
	return (memcmp(data, key.data, sizeof(xlic_license_key)) != 0);
}

#endif
