#ifndef __STRING_REF_HPP__
#define __STRING_REF_HPP__

#include <cstddef>
#include <cstring>
#include "config.h"

template<class T>
class basic_string_ref {
public:

	typedef typename std::basic_string<T>::iterator iterator;
	typedef typename std::basic_string<T>::const_iterator const_iterator;

	basic_string_ref() throw();

	template<size_t N> explicit basic_string_ref(const T (&string)[N]) throw();
	explicit basic_string_ref(const T * & string) throw();
	basic_string_ref(const T * string, const size_t length) throw();

	iterator begin() throw();
	const_iterator begin() const throw();

	const T * data() const throw();

	iterator end() throw();
	const_iterator end() const throw();

	size_t length() const throw();

	bool matches(const T * expression) const;

	bool operator==(const T * string) const throw();

	bool operator!=(const T * string) const throw();

private:

	const size_t m_length;
	const T * m_value;
};

typedef basic_string_ref<char> string_ref;

template<class T>
inline basic_string_ref<T>::basic_string_ref() throw()
: m_length(0)
, m_value(nullptr) {}

template<class T>
template<size_t N>
inline basic_string_ref<T>::basic_string_ref(const T (&string)[N]) throw()
: m_length(N)
, m_value(string) {}

template<class T>
inline basic_string_ref<T>::basic_string_ref(const T * & string) throw()
: m_length(strlen(string))
, m_value(string) {}

template<class T>
inline basic_string_ref<T>::basic_string_ref(const T * string, const size_t length) throw()
: m_length(length)
, m_value(string) {}

template<class T>
inline typename basic_string_ref<T>::iterator basic_string_ref<T>::begin() throw() {
	return iterator(m_value);
}

template<class T>
inline typename basic_string_ref<T>::const_iterator basic_string_ref<T>::begin() const throw() {
	return const_iterator(m_value);
}

template<class T>
inline const T * basic_string_ref<T>::data() const throw() {
	return m_value;
}

template<class T>
inline typename basic_string_ref<T>::iterator basic_string_ref<T>::end() throw() {
	return iterator(m_value + m_length);
}

template<class T>
inline typename basic_string_ref<T>::const_iterator basic_string_ref<T>::end() const throw() {
	return const_iterator(m_value + m_length);
}

template<class T>
inline size_t basic_string_ref<T>::length() const throw() {
	return m_length;
}

#if __cplusplus >= 201103L

#include <regex>

template<>
inline bool basic_string_ref<char>::matches(const char * expression) const {
	return std::regex_match(begin(), end(), std::regex(expression, std::regex_constants::basic));
}

template<>
inline bool basic_string_ref<wchar_t>::matches(const wchar_t * expression) const {
	return std::regex_match(begin(), end(), std::wregex(expression, std::regex_constants::basic));
}

#endif // __cplusplus >= 201103L

template<class T>
inline bool basic_string_ref<T>::operator==(const T * string) const throw() {
	return (m_length == strlen(string) && memcmp(m_value, string, m_length) == 0);
}

template<class T>
inline bool basic_string_ref<T>::operator!=(const T * string) const throw() {
	return (m_length != strlen(string) || memcmp(m_value, string, m_length) != 0);
}

template<class T>
inline std::basic_string<T> operator+(
	const basic_string_ref<T> & lhs,
	const basic_string_ref<T> & rhs
) {
	std::basic_string<T> out;
	out.reserve(lhs.length() + rhs.length());
	out.append(lhs.data(), lhs.length());
	out.append(rhs.data(), rhs.length());
	return out;
}

template<class T>
inline std::basic_string<T> operator+(
	const basic_string_ref<T> & lhs,
	const std::basic_string<T> & rhs
) {
	std::basic_string<T> out;
	out.reserve(lhs.length() + rhs.length());
	out.append(lhs.data(), lhs.length());
	out.append(rhs);
	return out;
}

template<class T, size_t N>
inline std::basic_string<T> operator+(
	const basic_string_ref<T> & lhs,
	const T (&rhs)[N]
) {
	std::basic_string<T> out;
	out.reserve(lhs.length() + N);
	out.append(lhs.data(), lhs.length());
	out.append(rhs, N);
	return out;
}

template<class T>
inline std::basic_string<T> operator+(
	const basic_string_ref<T> & lhs,
	const T * & rhs
) {
	const size_t length = strlen(rhs);
	std::basic_string<T> out;
	out.reserve(lhs.length() + length);
	out.append(lhs.data(), lhs.length());
	out.append(rhs, length);
	return out;
}

template<class T>
inline std::basic_string<T> operator+(
	const basic_string_ref<T> & lhs,
	const T rhs
) {
	std::basic_string<T> out;
	out.reserve(lhs.length() + 1);
	out.append(lhs.data(), lhs.length());
	out.append(&rhs, 1);
	return out;
}

template<class T>
inline std::basic_string<T> operator+(
	const std::basic_string<T> & lhs,
	const basic_string_ref<T> & rhs
) {
	std::basic_string<T> out;
	out.reserve(lhs.length() + rhs.length());
	out.append(lhs.data(), lhs.length());
	out.append(rhs.data(), rhs.length());
	return out;
}

template<class T, size_t N>
inline std::basic_string<T> operator+(
	const T (&lhs)[N],
	const basic_string_ref<T> & rhs
) {
	std::basic_string<T> out;
	out.reserve(N + rhs.length());
	out.append(lhs, N);
	out.append(rhs.data(), rhs.length());
	return out;
}

template<class T>
inline std::basic_string<T> operator+(
	const T * & lhs,
	const basic_string_ref<T> & rhs
) {
	const size_t length = strlen(lhs);
	std::basic_string<T> out;
	out.reserve(length + rhs.length());
	out.append(lhs, length);
	out.append(rhs.data(), rhs.length());
	return out;
}

template<class T>
inline std::basic_string<T> operator+(
	const T lhs,
	const basic_string_ref<T> & rhs
) {
	std::basic_string<T> out;
	out.reserve(1 + rhs.length());
	out.append(&lhs, 1);
	out.append(rhs.data(), rhs.length());
	return out;
}

#endif // __STRING_REF_HPP__
