#include <cinttypes>
#include <iostream>
#include <unistd.h>
#include <xorcom/licensing.h>
#include "hex.hpp"
#include "base64.hpp"
#include "util.hpp"
#include "globals.hpp"

#ifndef XLIC_VERSION
#define XLIC_VERSION __DATE__ " " __TIME__
#endif

inline static int print_usage() {
	puts(
		"Usage: xlic [COMMAND] <FLAG>... [ARGUMENT]\n"
		"Xorcom product licensing " XLIC_VERSION ".\n"
		"Copyright (C) 2015 Xorcom Ltd.\n"
		"\n"
		"Available COMMANDs:\n"
		"  find         Display license information where ARGUMENT is a product name.\n"
		"  list         Display all licenses currently active.\n"
		"  register     Attempt to register license where ARGUMENT is a license key.\n"
		"  remove       Remove license where ARGUMENT is a license key.\n"
		"  sysinfo      Print system's MAC and serial number.\n"
		"  type         Determine the license type where ARGUMENT is a license key.\n"
		"  update       Contact licensing server and update licensing information.\n"
		"\n"
		"Optional FLAGs:\n"
		"  -s <SERIAL>  (Virtual licenses only) Use SERIAL number instead for registration and updates.\n"
		"               Note: This will not override obtaining a serial number from the dongle.\n"
		"  -o <FILE>    Use FILE instead for license storage.\n"
		"  -u <URL>     Use URL instead to contact the licensing server.\n"
	);
	return EINVAL;
}

inline static int print_sysinfo() {
	std::cout <<
		"MAC               : " << mac_address() << "\n" <<
		"Serial            : " << dongle_serial_number() << "\n" <<
		"Serial File       : " << file_serial_number() << "\n";
	return EXIT_SUCCESS;
}

inline static std::string human_readable(size_t value) {
	char buffer[200]; // probably enough? (per manpage example)
	const time_t timestamp = value;
	return std::string(buffer, strftime(buffer, sizeof(buffer), "%F %T", localtime(&timestamp)));
}

inline static void print_license(const xlic_license & license) {
	printf(
		"Type              : %s\n"
		"Name              : %s\n"
		"Displayname       : %s\n"
		"Last Registration : %s\n"
		"Expires           : %s\n"
		"License Key       : %s\n"
		"Signature         : %s\n\n",
		(license.type == XLIC_PHYSICAL_LICENSE) ? "Physical" : "Virtual",
		license.name,
		license.description,
		human_readable(license.generated / 1000000).data(),
		(license.expires > 0) ? human_readable(license.expires).data() : "Never",
		base64::encode(&license.key, sizeof(xlic_license_key)).data(),
		base64::encode(&license.signature, sizeof(xlic_signature)).data()
	);
}

inline static int list_licenses(const char * filename) {
	xlic_license_list * list = nullptr;
	const int res = xlic_list_licenses(&list, filename);
	if (res == EXIT_SUCCESS && list) {
		if (list->count > 0) {
			printf(
				"Licenses          : %zu\n"
				"Public Key        : %s\n\n",
				list->count,
				base64::encode(&list->public_key, sizeof(xlic_public_key)).data()
			);
		} else {
			puts("No license file found or license file is empty");
		}
		for (size_t i = 0, end = list->count; i < end; i++) {
			print_license(list->licenses[i]);
		}
		xlic_destroy_license_list(list);
	}
	return res;
}

inline static int find_license(
	const char * filename,
	const char * product_name
) {
	xlic_license * license = nullptr;
	const int res = xlic_find_product_license(&license, filename, product_name);
	if (res == EXIT_SUCCESS && license) {
		print_license(*license);
		xlic_destroy_license(license);
	} else {
		fprintf(stderr, "No such product is licensed on this system\n");
	}
	return res;
}

inline static int register_license(
	const char * filename,
	const char * url,
	const char * license_key,
	const char * serial_number
) {
	const int res = xlic_register_license(filename, url, license_key, serial_number);
	if (res == EXIT_SUCCESS) {
		puts("Registration successful");
	}
	return res;
}

inline static int remove_license(
	const char * filename,
	const char * url,
	const char * license_key
) {
	const int res = xlic_remove_license(filename, url, license_key);
	if (res == EXIT_SUCCESS) {
		puts("License successfully removed");
	}
	return res;
}

inline static int license_type(const char * license_key) {
	switch (xlic_license_key_type(license_key)) {
		case XLIC_INVALID_LICENSE: {
			puts("License key is invalid");
			return EINVAL;
		}
		case XLIC_VIRTUAL_LICENSE: {
			puts("License is virtual");
			break;
		}
		case XLIC_PHYSICAL_LICENSE: {
			puts("License is physical");
			break;
		}
	}
	return EXIT_SUCCESS;
}

int main(int argc, char ** argv) {
	const char * url = XLIC_LICENSE_SERVER;
	const char * filename = XLIC_LICENSE_FILE;
	const char * serial_number = "";
	int opt = 0;
	while ((opt = getopt(argc, argv, "s:o:u:")) != -1) {
		switch (opt) {
			case 's': { serial_number = optarg; break; }
			case 'o': { filename = optarg; break; }
			case 'u': { url = optarg; break; }
			case '?': {
				printf("? %c\n", optopt);
				return print_usage();
			}
			default: { return EINVAL; }
		}
	}
	argc -= optind;
	argv += optind;
	if (argc > 0) {
		if (strcmp(argv[0], "find") == 0) {
			if (filename && argc > 1) {
				return find_license(filename, argv[1]);
			}
		} else if (strcmp(argv[0], "list") == 0) {
			if (filename) {
				return list_licenses(filename);
			}
		} else if (strcmp(argv[0], "register") == 0) {
			if (filename && url && argc > 1) {
				return register_license(filename, url, argv[1], serial_number);
			}
		} else if (strcmp(argv[0], "remove") == 0) {
			if (filename && url && argc > 1) {
				return remove_license(filename, url, argv[1]);
			}
		} else if (strcmp(argv[0], "sysinfo") == 0) {
			return print_sysinfo();
		} else if (strcmp(argv[0], "type") == 0) {
			if (argc > 1) {
				return license_type(argv[1]);
			}
		} else if (strcmp(argv[0], "update") == 0) {
			if (filename && url) {
				return xlic_update_licenses(filename, url);
			}
		}
	}
	return print_usage();
}
